/* program to print mass spectra given formula */

#define CUTOFF 5e-5

#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>

typedef struct {int m;  float fr; } isotope;
typedef struct {char *sym;  int niso;  isotope *p; } element;
typedef struct {int atno;  int count; } atom;
typedef struct {int mass;  float intens; } peak;

element el[] =
  {
#include "eltab"
  }; 

isotope iso[] =
{           
#include "isotab"
};

#define ADDBASE 100              /* above the natural elements */
#define MAXADD 10                /* space for user-defined 'elements' */
element addel[MAXADD];           /* eg isotope enriched ones, etc */
isotope addiso[5 * MAXADD];
int eadd=0, iadd=0;


#define MAXAT  50
atom atoms[MAXAT];
int natoms;

int nel = sizeof(el) / sizeof(element);
int niso = sizeof(iso) / sizeof(isotope);


int atno(char *s)       /* return at # or 0 if s not vaalid element symbol */
{
   int i;

   for (i=0;  i<nel;  i++)
   {  if (0 == strcmp(el[i].sym, s)) return(i + 1);
   }
   for (i=0;  i<eadd;  i++)
   {  if (0 == strcmp(addel[i].sym, s)) return(i + ADDBASE);
   }                 /* return a high 'atomic #' on user-defined elements */
   return(0);
}

void setpointers()      /* set pointers in el entries to start of isotopes */
{                       /* for that element in the iso table */
   int i;
   isotope *p;

   p = iso;
   for (i=0;  i<nel;  i++)
   {  el[i].p = p;
      p += el[i].niso;
   }
}


isotope imin(int atno)        /* isotope of lowest mass of element atno */
{  if (atno < ADDBASE)
      return (*(el[atno-1].p));  /* element 1 is in table entry 0, etc. */
   else return(*(addel[atno-ADDBASE].p));
}

isotope imax(int atno)        /* isotope of highest mass of element atno */
{  if (atno < ADDBASE)
      return( *(el[atno - 1].p + el[atno -1].niso - 1));
   else return(*(addel[atno-ADDBASE].p + addel[atno -ADDBASE].niso - 1));
}


main(int argc, char *argv[])
{  int i, j, m, q, nold, nnew, oldmin, oldmax, newmin, newmax;
   register int k;
   int ii, ix, fraction, ns;
   char b[81], stars[71];
   float fr, maxintens, sumintens;
   isotope s;
   element e;
   peak *old, *new;
   register peak *pp;

   fraction = (argc > 1 && 0 == strcmp(argv[1], "-f"));
         /* normalize to 100 max, or print fractions on -f cmd switch */

   setpointers();
   printf(" Enter a formula, or Q to quit, or E to define an extra element\n\n");

   while(1)
   {  printf(":");
      fgets(b, 80, stdin);
      b[80] = 0;
      squob(b);
      for(k=0;  k<strlen(b);  k++)
      {  if(b[k] == '\n' || b[k] == '\r')
         {  b[k] = 0;
            break;
         }
      }

      if (b[0] == 'q' || b[0] == 'Q') exit(0);

      if ((b[0] == 'e' || b[0] == 'E') && strlen(b) == 1)
      {  addelement();
         continue;
      }

      natoms = 0;
      k = formula(b);
      if (k == 0) continue;

      old = (peak *)malloc(sizeof(peak));
      old->mass = 0;
      old->intens = 1;
      nold = 1;
      oldmin = oldmax = 0;

      for (i=0;  i<natoms;  i++)
      {  for(j=0;  j<atoms[i].count;  j++)
         {  s = imin(atoms[i].atno);
            newmin = oldmin + s.m;
            s = imax(atoms[i].atno);
            newmax = oldmax + s.m;
            nnew = newmax - newmin + 1;
            new = (peak *)malloc(nnew * sizeof(peak));
            if(new == NULL)
            {  printf("\nOut of Memory!\n");
               exit(1);
            }
            for(k=0;  k<nnew;  k++)
            {  
               new[k].intens = 0;
            }
            if (atoms[i].atno < ADDBASE) e = el[atoms[i].atno - 1];
            else e = addel[atoms[i].atno - ADDBASE];

            for(k=0;  k<e.niso;  k++)
            {  m = (k + e.p)->m;
               fr = (k + e.p)->fr;
               for (q=0;  q<nold;  q++)
               {  ix = m + old[q].mass - newmin;
#ifdef DEBUG
   printf(" %d", ix);
#endif
                  new[ix].mass = m + old[q].mass;
                  new[ix].intens += fr * old[q].intens;
               }
            }
                               /* normalize to maximum intensity of 1.0 */
             maxintens = 0;
             for(ii=0;  ii<nnew;  ii++)
                if (new[ii].intens > maxintens) maxintens = new[ii].intens;
             for(ii=0;  ii<nnew;  ii++) new[ii].intens /= maxintens;

                                        /* throw away very small peaks */
             for (ii=0;  ii<nnew;  ii++)
             {  if(new[ii].intens < CUTOFF)
                {  for (k=ii, pp = new + ii;  k<nnew-1;  k++, pp++) 
                   {  pp->mass = (pp+1)->mass;
                      pp->intens = (pp+1)->intens;   /* inner loop...*/
                   }                                 /* avoid structure copy */
                   nnew--;
                   ii--;
                }
             }

#ifdef DEBUG
   printf("\n");
   for(ii=0;  ii<nnew;  ii++)
   {  if (new[ii].intens == 0) continue;
      printf("%3d%10.4g\n", new[ii].mass, new[ii].intens);
   }
      printf("\n");
#endif

            free(old);   
            old = new;
            nold = nnew;
            oldmin = newmin;
            oldmax = newmax;
         }
      }

      maxintens = sumintens = 0;
      for (ii=0;  ii<nnew;  ii++)
      {  sumintens += new[ii].intens;
         if (new[ii].intens > maxintens) maxintens = new[ii].intens;
      }
      for (ii=0;  ii<nnew;  ii++)
      {  if (fraction) new[ii].intens /= sumintens;
         else new[ii].intens *= 100;  /* they are already normalized to max=1 */
      }
      if (fraction) maxintens /= sumintens;
      else maxintens *= 100;

      for(ii=0;  ii<nnew;  ii++)
      {  ns = .5 + 60 * new[ii].intens / maxintens;
         for (k=0;  k<ns;  k++) stars[k] = '*';
         stars[ns] = 0;
         printf(fraction? "%5d%8.4f  |%s\n" : "%5d%8.2f  |%s\n", 
                new[ii].mass, new[ii].intens, stars);
      }
      printf("\n");
      free(new);
   }
}


int formula(char *in)      /* determine if input is valid formula */
{                          /* return 1 if so, 0 if not call foutput() */
   int a, i, n, state;     /* to handle (element, count) pairs */
   char k, s[3];

   state = 0;

   while(0 != (k = *(in++)))
   {  if (k == '\n' || k == '\r') continue;  /* junk at end from fgets ? */
      switch(state)
      {  case 0:
            if(isupper(k))
            {  s[0] = k;
               s[1] = 0;
               state = 1;
            }
            else goto error;
            break;

         case 1:
            if (isdigit(k))
            {  if(0 == (a = atno(s))) goto error;
               n = k - '0';
               state = 2;
            }
            else if(islower(k))
            {  s[1] = k;
               s[2] = 0;
               if(0 == (a = atno(s))) goto error;
               state = 3;
            }
            else if(isupper(k))
            {  if(0 == (a = atno(s))) goto error;
               n = 1;
               foutput(a, n);
               state = 1;
               s[0] = k;
               s[1] = 0;
            }
            else goto error;
            break;

         case 2:
            if (isdigit(k)) n = 10 * n + k - '0';
            else if(isupper(k))
            {  foutput(a, n);
               state = 1;
               s[0] = k;
               s[1] = 0;
            }
            else goto error;
            break;

         case 3:
            if (isdigit(k))
            {  n = k - '0';
               state = 2;
            }
            else if (isupper(k))
            {  if(0 == (a = atno(s))) goto error;
               n = 1;
               foutput(a, n);
               state = 1;
               s[0] = k;
               s[1] = 0;
            }
            else goto error;
            break;

      }  /* end of case */
   }     /* end of while */

   if (state == 1 || state == 3) n = 1;
   if (state != 0)
   {  if(0 == (a = atno(s))) goto error;
      foutput(a, n);
      return(1);
   }

error:
   printf("Bad formula\n\n");
   return(0);
}


foutput(int a, int n)
{
#ifdef DEBUG
   isotope low, high;
#endif
   atoms[natoms].atno = a;
   atoms[natoms].count = n;
   natoms++;
   if (natoms >= MAXAT)
   {  printf("Formula too long\n");
      exit(1);
   }
#ifdef DEBUG
   low = imin(a);
   high = imax(a);
   printf("%3d%5d   min: %d   max: %d\n", a, n, low.m, high.m);
#endif
}


addelement()            /* user-defined element */
{
   int i, j, k, first;
   float f, sumpc;
   char b2[81], b[81], *p;

   while(1)
   {  printf("Enter symbol for element: ");
      fgets(b, 80, stdin);
      for(i=0;  i<strlen(b);  i++) if(b[i] == '\n' || b[i] == '\r') b[i] = ' ';
      squob(b);
      b[2] = 0;
      if (!isupper(b[0]) || (b[1] != 0 && (!islower(b[1]))))
      {  printf("Bad symbol\n\n");
         continue;
      }
      if(0 == atno(b))     /* not already known symbol */
      {  addel[eadd].sym = (char *)malloc(3);
         strcpy(addel[eadd].sym, b);
         break;
      }
      printf("This symbol is already in use.\n\n");
   }
   printf("\nEnter mass - percent abundance pairs, one pair per line, with space between.\n");
   printf("Enter an empty line to finish.\n");

startiso:
   addel[eadd].niso = 0;
   first = 1;
   sumpc = 0;
   while(1)
   {  printf(">");
      fgets(b, 80, stdin);
      if (strlen(b) <= 1)
      {  if(sumpc > 99.5 && sumpc < 100.5) break;
         else if(sumpc <= 99.5)
         {  printf("Sum of isotope percentages is %.1f, press Y if ok, N to enter more:",
                                      sumpc);
            fgets(b2, 80, stdin);
            if (b2[0] == 'y' || b2[0] == 'Y') break;
            else continue;
         }
         else
         {  printf("Sum of isotope percentages is %.1f, press Y if ok, N to start over:",
                                      sumpc);
            fgets(b2, 80, stdin);
            if (b2[0] == 'y' || b2[0] == 'Y') break;
            else goto startiso;
         }
      }
      j = strtol(b, &p, 10);

      if (j > 0 && j < 300) addiso[iadd].m = j;
      else if(j <= 0)
      {  printf("Negative or zero mass. Re-enter line.");
         continue;
      }
      else
      {  printf("Mass > 300. Type Y to confirm, or N to re-enter line. ");
         fgets(b2, 80, stdin);
         k = b2[0];
         if (k != 'y' && k != 'Y') continue;
         addiso[iadd].m = j;
      }


      f = atof(p);
      if (f > 0 && f <= 100) 
      {  addiso[iadd].fr = f / 100;
         sumpc += f;
      }
      else
      {  printf("Impossible percentage. Re-enter line\n");
         continue;
      }

      addel[eadd].niso++;
      if(first)
      {  addel[eadd].p = addiso + iadd;
         first = 0;
      }
      iadd++;
   }
   eadd++;
}
      


squob(s)            /* squeeze out the blanks in s and return the length */
    char *s;        /* of the resultant string not counting terminating null*/
{   int i,j;
    char c;

    i=j=0;
    while ( (c=s[i++]) != '\0')     if (c != ' ') s[j++]=c;
    s[j]='\0';
    return(j);
}

